##############################################################################
#
# Copyright (c) 2013 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import json
import logging
import pprint
import unittest
import tempfile
import shutil
import socket
import textwrap
import sys
import os
import pkg_resources

from contextlib import contextmanager
from mock import patch, create_autospec
import mock
from slapos.util import sqlite_connect, bytes2str, UndefinedSerializationError
from slapos.slap.slap import DEFAULT_SOFTWARE_TYPE

import slapos.cli.console
import slapos.cli.entry
import slapos.cli.info
import slapos.cli.list
import slapos.cli.computer_info
import slapos.cli.computer_list
import slapos.cli.computer_token
import slapos.cli.supervisorctl
import slapos.cli.request
from slapos.cli.proxy_show import do_show, StringIO
from slapos.cli.cache_binarysr import do_lookup as cache_binarysr_do_lookup
from slapos.cli.cache_url import do_lookup as cache_url_do_lookup
from slapos.cli.cache_pypi import do_lookup as cache_pypi_do_lookup
from slapos.client import ClientConfig
from slapos.slap import SoftwareProductCollection
import slapos.grid.svcbackend
import slapos.proxy
import slapos.slap

import six
import supervisor.supervisorctl

signature_certificate_list = """-----BEGIN CERTIFICATE-----
MIIB9jCCAV+gAwIBAgIJAKRvzcy7OH0UMA0GCSqGSIb3DQEBBQUAMBMxETAPBgNV
BAMMCENPTVAtNzcyMCAXDTEyMDgxMDE1NDI1MVoYDzIxMTIwNzE3MTU0MjUxWjAT
MREwDwYDVQQDDAhDT01QLTc3MjCBnzANBgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEA
o7aipd6MbnuGDeR1UJUjuMLQUariAyQ2l2ZDS6TfOwjHiPw/mhzkielgk73kqN7A
sUREx41eTcYCXzTq3WP3xCLE4LxLg1eIhd4nwNHj8H18xR9aP0AGjo4UFl5BOMa1
mwoyBt3VtfGtUmb8whpeJgHhqrPPxLoON+i6fIbXDaUCAwEAAaNQME4wHQYDVR0O
BBYEFEfjy3OopT2lOksKmKBNHTJE2hFlMB8GA1UdIwQYMBaAFEfjy3OopT2lOksK
mKBNHTJE2hFlMAwGA1UdEwQFMAMBAf8wDQYJKoZIhvcNAQEFBQADgYEAaNRx6YN2
M/p3R8/xS6zvH1EqJ3FFD7XeAQ52WuQnKSREzuw0dsw12ClxjcHiQEFioyTiTtjs
5pW18Ry5Ie7iFK4cQMerZwWPxBodEbAteYlRsI6kePV7Gf735Y1RpuN8qZ2sYL6e
x2IMeSwJ82BpdEI5niXxB+iT0HxhmR+XaMI=
-----END CERTIFICATE-----
"""

def raiseNotFoundError(*args, **kwargs):
  raise slapos.slap.NotFoundError()

class CliMixin(unittest.TestCase):
  def setUp(self):
    slap = slapos.slap.slap()
    self.logger = create_autospec(logging.Logger)
    self.local = {'slap': slap, 'product': SoftwareProductCollection(self.logger, slap)}
    self.conf = create_autospec(ClientConfig)
    self.sign_cert_list = signature_certificate_list

class TestCliCacheBinarySr(CliMixin):

  test_url = "https://lab.nexedi.com/nexedi/slapos/raw/1.0.102/software/slaprunner/software.cfg"
  def test_cached_binary(self):
    self.assertEqual(0, cache_binarysr_do_lookup(
        self.logger,
        cache_dir="http://dir.shacache.org",
        cache_url="http://shacache.org",
        software_url=self.test_url,
        signature_certificate_list=self.sign_cert_list))

    self.logger.info.assert_any_call('Software URL: %s',
            u'https://lab.nexedi.com/nexedi/slapos/raw/1.0.102/software/slaprunner/software.cfg')
    self.logger.info.assert_any_call('MD5:          %s', 'cccdc51a07e8c575c880f2d70dd4d458')
    self.logger.info.assert_any_call(u'---------------------------------------------------------------------')
    self.logger.info.assert_any_call(u'    multiarch     distribution version    id   compatible? verified? ')
    self.logger.info.assert_any_call(u'---------------------------------------------------------------------')
    self.logger.info.assert_any_call(u' x86_64-linux-gnu CentOS Linux 7.5.1804  Core       no        yes    ')
    self.logger.info.assert_any_call(u' x86_64-linux-gnu    Ubuntu     18.04   bionic      no        yes    ')
    # Omit some lines as it may fail depending of the OS
    self.logger.info.assert_any_call(u'---------------------------------------------------------------------')

  def test_uncached_binary(self):
    self.assertEqual(1, cache_binarysr_do_lookup(
        self.logger,
        cache_dir="http://dir.shacache.org",
        cache_url="http://shacache.org",
        software_url="this_is_uncached_url",
        signature_certificate_list=self.sign_cert_list))

    self.logger.critical.assert_any_call(
        'Error while looking object %s', 'this_is_uncached_url', exc_info=True)

  def test_bad_cache_dir(self):
    self.assertEqual(1, cache_binarysr_do_lookup(
        self.logger,
        cache_dir="http://xxx.shacache.org",
        cache_url="http://shacache.org",
        software_url=self.test_url,
        signature_certificate_list=self.sign_cert_list))

    self.logger.critical.assert_any_call(
        'Error while looking object %s',
        'https://lab.nexedi.com/nexedi/slapos/raw/1.0.102/software/slaprunner/software.cfg',
        exc_info=True)

  def test_unverified_signature(self):
      with mock.patch(
              'slapos.grid.networkcache.machine_info_tuple',
              return_value=('x86_64-linux-gnu', ('debian', '8.10', ''))):
          self.assertEqual(0, cache_binarysr_do_lookup(
              self.logger,
              cache_dir="http://dir.shacache.org",
              cache_url="http://shacache.org",
              software_url=self.test_url,
              signature_certificate_list=""))

      self.logger.info.assert_any_call('Software URL: %s',
              u'https://lab.nexedi.com/nexedi/slapos/raw/1.0.102/software/slaprunner/software.cfg')
      self.logger.info.assert_any_call('MD5:          %s', 'cccdc51a07e8c575c880f2d70dd4d458')
      self.logger.info.assert_any_call(u'---------------------------------------------------------------------')
      self.logger.info.assert_any_call(u'    multiarch     distribution version    id   compatible? verified? ')
      self.logger.info.assert_any_call(u'---------------------------------------------------------------------')
      self.logger.info.assert_any_call(u' x86_64-linux-gnu CentOS Linux 7.5.1804  Core       no         no    ')
      self.logger.info.assert_any_call(u' x86_64-linux-gnu    Ubuntu     18.04   bionic      no         no    ')
      self.logger.info.assert_any_call(u' x86_64-linux-gnu    debian      8.10              yes         no    ')
      self.logger.info.assert_any_call(u'---------------------------------------------------------------------')


class TestCliCacheUrl(CliMixin):

  test_url = "https://ftp.gnu.org/gnu/aspell/aspell-0.60.7.tar.gz"
  cache_dir = "http://dir.shacache.org"
  def test_cached_url(self):
    self.assertEqual(0, cache_url_do_lookup(
        self.logger,
        cache_dir=self.cache_dir,
        cache_url="http://shacache.org",
        url=self.test_url,
        signature_certificate_list=""))


    self.logger.info.assert_any_call('Software source URL: %s', self.test_url)
    self.logger.info.assert_any_call('SHADIR URL: %s/%s\n', self.cache_dir, "file-urlmd5:f213fcd8e97aa729f685b8cb71b976a7")
    self.logger.info.assert_any_call(u'---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
    self.logger.info.assert_any_call(u'                         url                                                                                      sha512                                                              signed ')
    self.logger.info.assert_any_call(u'---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
    self.logger.info.assert_any_call(u' https://ftp.gnu.org/gnu/aspell/aspell-0.60.7.tar.gz 6f5fcd1c29164ee18f205594b66f382b51d19b17686293a931ca92c1442d3f7228627ca7d604d860551d0d367ac34dfb2ae34170a844f51e84e390fb1edc4535 False  ')
    self.logger.info.assert_any_call(u'---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')

  def test_cached_signed_url(self):
    self.assertEqual(0, cache_url_do_lookup(
        self.logger,
        cache_dir=self.cache_dir,
        cache_url="http://shacache.org",
        url=self.test_url,
        signature_certificate_list=signature_certificate_list))


    self.logger.info.assert_any_call('Software source URL: %s', self.test_url)
    self.logger.info.assert_any_call('SHADIR URL: %s/%s\n', self.cache_dir, "file-urlmd5:f213fcd8e97aa729f685b8cb71b976a7")
    self.logger.info.assert_any_call(u'---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
    self.logger.info.assert_any_call(u'                         url                                                                                      sha512                                                              signed ')
    self.logger.info.assert_any_call(u'---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')
    self.logger.info.assert_any_call(u' https://ftp.gnu.org/gnu/aspell/aspell-0.60.7.tar.gz 6f5fcd1c29164ee18f205594b66f382b51d19b17686293a931ca92c1442d3f7228627ca7d604d860551d0d367ac34dfb2ae34170a844f51e84e390fb1edc4535  True  ')
    self.logger.info.assert_any_call(u'---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------')

  def test_uncached_url(self):
    self.assertEqual(1, cache_url_do_lookup(
        self.logger,
        cache_dir="http://dir.shacache.org",
        cache_url="http://shacache.org",
        url="this_is_uncached_url",
        signature_certificate_list=""))

    self.logger.info.assert_any_call('Object not found in cache.')

  def test_bad_cache_dir(self):
    self.assertEqual(1, cache_url_do_lookup(
        self.logger,
        cache_dir="http://xxx.shacache.org",
        cache_url="http://shacache.org",
        url=self.test_url,
        signature_certificate_list=""))

    self.logger.critical.assert_any_call('Error while looking object %s', self.test_url, exc_info=True)

class TestCliCachePypi(CliMixin):

  egg_name = "pytz"
  egg_version = "2016.10"
  cache_dir = "http://dir.shacache.org"
  def test_cached_pypi(self):
    self.assertEqual(0, cache_pypi_do_lookup(
        self.logger,
        cache_dir=self.cache_dir,
        cache_url="http://shacache.org",
        name=self.egg_name,
        version=self.egg_version,
        signature_certificate_list=""))


    self.logger.info.assert_any_call('Python egg %s version %s', self.egg_name, self.egg_version)
    self.logger.info.assert_any_call('SHADIR URL: %s/%s\n', self.cache_dir, 'pypi:{}={}'.format(self.egg_name, self.egg_version))
    self.logger.info.assert_any_call('----------------------------------------------------------------------------------------------------------------------------------------------------------------')
    self.logger.info.assert_any_call('        basename                                                                     sha512                                                              signed ')
    self.logger.info.assert_any_call('----------------------------------------------------------------------------------------------------------------------------------------------------------------')
    self.logger.info.assert_any_call(' pytz-2016.10-py2.7.egg e072d146c42cb2efde946eef1d37ce6f8cb8eec6f6f928f6d57bb5312578bfa0031dcdbd816318015d765886bb64c02e1772adf7309142bc80324a4155b4ae8b False  ')
    self.logger.info.assert_any_call(' pytz-2016.10-py2.7.egg e072d146c42cb2efde946eef1d37ce6f8cb8eec6f6f928f6d57bb5312578bfa0031dcdbd816318015d765886bb64c02e1772adf7309142bc80324a4155b4ae8b False  ')
    self.logger.info.assert_any_call('----------------------------------------------------------------------------------------------------------------------------------------------------------------')


  def test_cached_signed_url(self):
    self.assertEqual(0, cache_pypi_do_lookup(
        self.logger,
        cache_dir=self.cache_dir,
        cache_url="http://shacache.org",
        name=self.egg_name,
        version=self.egg_version,
        signature_certificate_list=signature_certificate_list))


    self.logger.info.assert_any_call('Python egg %s version %s', self.egg_name, self.egg_version)
    self.logger.info.assert_any_call('SHADIR URL: %s/%s\n', self.cache_dir, 'pypi:{}={}'.format(self.egg_name, self.egg_version))
    self.logger.info.assert_any_call('----------------------------------------------------------------------------------------------------------------------------------------------------------------')
    self.logger.info.assert_any_call('        basename                                                                     sha512                                                              signed ')
    self.logger.info.assert_any_call('----------------------------------------------------------------------------------------------------------------------------------------------------------------')
    self.logger.info.assert_any_call(' pytz-2016.10-py2.7.egg e072d146c42cb2efde946eef1d37ce6f8cb8eec6f6f928f6d57bb5312578bfa0031dcdbd816318015d765886bb64c02e1772adf7309142bc80324a4155b4ae8b  True  ')
    self.logger.info.assert_any_call(' pytz-2016.10-py2.7.egg e072d146c42cb2efde946eef1d37ce6f8cb8eec6f6f928f6d57bb5312578bfa0031dcdbd816318015d765886bb64c02e1772adf7309142bc80324a4155b4ae8b False  ')
    self.logger.info.assert_any_call('----------------------------------------------------------------------------------------------------------------------------------------------------------------')

  def test_uncached_url(self):
    self.assertEqual(1, cache_pypi_do_lookup(
        self.logger,
        cache_dir="http://dir.shacache.org",
        cache_url="http://shacache.org",
        name="not-existing-egg",
        version=self.egg_version,
        signature_certificate_list=""))

    self.logger.info.assert_any_call('Object not found in cache.')

  def test_bad_cache_dir(self):
    self.assertEqual(1, cache_pypi_do_lookup(
        self.logger,
        cache_dir="http://xxx.shacache.org",
        cache_url="http://shacache.org",
        name=self.egg_name,
        version=self.egg_version,
        signature_certificate_list=""))

    self.logger.critical.assert_any_call('Error while looking egg %s version %s', self.egg_name, self.egg_version, exc_info=True)

class TestCliProxy(CliMixin):
  def test_generateSoftwareProductListFromString(self):
    """
    Test that generateSoftwareProductListFromString correctly parses a parameter
    coming from the configuration file.
    """
    software_product_list_string = """
product1 url1
product2 url2"""
    software_release_url_list = {
        'product1': 'url1',
        'product2': 'url2',
    }
    self.assertEqual(
        slapos.proxy._generateSoftwareProductListFromString(
            software_product_list_string),
        software_release_url_list
    )

  def test_generateSoftwareProductListFromString_emptyString(self):
    self.assertEqual(
        slapos.proxy._generateSoftwareProductListFromString(''),
        {}
    )

class TestCliProxyShow(CliMixin):
  def setUp(self):
    super(TestCliProxyShow, self).setUp()
    self.db_file = tempfile.NamedTemporaryFile(delete=False, suffix='.db')
    self.db_file.close()
    self.conf.database_uri = self.db_file.name
    self.conf.logger = self.logger

    # load database
    schema = bytes2str(pkg_resources.resource_string(
        'slapos.tests',
        os.path.join('test_slapproxy', 'database_dump_version_current.sql')))
    db = sqlite_connect(self.db_file.name)
    db.cursor().executescript(schema)
    db.commit()

    # by default we simulate being invoked with "show all" arguments
    self.conf.computers = True
    self.conf.software = True
    self.conf.partitions = True
    self.conf.slaves = True
    self.conf.params = True
    self.conf.network = True

  def tearDown(self):
    super(TestCliProxyShow, self).tearDown()
    os.remove(self.db_file.name)

  def test_proxy_show(self):
    logger = create_autospec(logging.Logger)
    with mock.patch(
            'slapos.cli.proxy_show.logging.getLogger',
            return_value=logger):
        do_show(self.conf)

    # installed softwares are listed
    logger.info.assert_any_call(
        '      /srv/slapgrid/slappart8/srv/runner/project/slapos/software/erp5/software.cfg          slaprunner        available    287375f0cba269902ba1bc50242839d7 ')

    # instance parameters are listed
    # _ parameter is json formatted
    logger.info.assert_any_call(
        '    %s = %s',
        '_',
        '{\n  "monitor-base-url": "",\n  "url": "memcached://10.0.30.235:2003/"\n}')

    # other parameters are displayed as simple string
    logger.info.assert_any_call(
        '    %s = %s',
        'url',
        'http://10.0.30.235:4444/wd/hub')

    # if _ cannot be decoded as json, it is displayed "as is"
    logger.info.assert_any_call(
        '    %s = %s',
        '_',
        u'Ahah this is not json \U0001f61c ')
    # Nothing was output on application logger, because this command uses
    # its own logger.
    self.logger.info.assert_not_called()

  def test_proxy_show_displays_on_stdout(self):
    # we patch logging to make sure our messages are outputed, even with test
    # runners like pytest which allows disabling output
    with patch.object(sys, 'stdout', StringIO()) as stdout, \
        patch.object(sys, 'stderr', StringIO()) as stderr, \
        patch('logging.Logger.isEnabledFor', returned_value=True):
      do_show(self.conf)

    # 287375f0cba269902ba1bc50242839d7 is the hash of an installed software
    # in our setup database
    self.assertIn('287375f0cba269902ba1bc50242839d7', stdout.getvalue())
    self.assertEqual('', stderr.getvalue())

  def test_proxy_show_use_pager(self):
    # we patch logging to make sure our messages are outputed, even with test
    # runners like pytest which allows disabling output
    with patch.object(sys, 'stdout', StringIO()) as stdout, \
        patch.object(sys, 'stderr', StringIO()) as stderr, \
        patch('logging.Logger.isEnabledFor', returned_value=True):
      stdout.isatty = lambda *args: True

      # use a pager that just output to a file.
      tmp = tempfile.NamedTemporaryFile(delete=False)
      self.addCleanup(os.unlink, tmp.name)
      os.environ['PAGER'] = 'cat > {}'.format(tmp.name)

      do_show(self.conf)

    self.assertEqual('', stdout.getvalue())
    self.assertEqual('', stderr.getvalue())
    # our pager was set to output to this temporary file
    with open(tmp.name, 'r') as f:
      self.assertIn('287375f0cba269902ba1bc50242839d7', f.read())


class TestCliBoot(CliMixin):
  def test_node_boot(self):
    # initialize instance root, with a timestamp in a partition
    temp_dir = tempfile.mkdtemp()
    self.addCleanup(shutil.rmtree, temp_dir)
    instance_root = os.path.join(temp_dir, 'instance')
    partition_base_name = 'partition'
    os.mkdir(instance_root)
    os.mkdir(os.path.join(instance_root, partition_base_name + '1'))
    timestamp = os.path.join(
        instance_root, partition_base_name + '1', '.timestamp')
    requested_state_path = os.path.join(instance_root,
                                        partition_base_name + '1',
                                        '.requested_state')
    with open(timestamp, 'w') as f:
      f.write("1578552471")
    with open(requested_state_path, 'w') as f:
      f.write("started")

    # make a config file using this instance root
    with tempfile.NamedTemporaryFile(mode='w') as slapos_conf:
      slapos_conf.write(
          textwrap.dedent(
              """\
              [slapos]
              instance_root = {instance_root}
              master_url = https://slap.vifib.com/

              [slapformat]
              partition_base_name = {partition_base_name}
              interface_name = interface_name_from_config
              """).format(**locals()))
      slapos_conf.flush()

      # run slapos node boot
      app = slapos.cli.entry.SlapOSApp()
      with patch('slapos.cli.boot.check_root_user', return_value=True) as check_root_user,\
          patch('slapos.cli.boot.SlapOSApp') as SlapOSApp,\
          patch('slapos.cli.boot.ConfigCommand.config_path', return_value=slapos_conf.name), \
          patch(
              'slapos.cli.boot.netifaces.ifaddresses',
              return_value={socket.AF_INET6: ({'addr': '2000::1'},),},) as ifaddresses,\
          patch('slapos.cli.boot._startComputerPartition', return_value=None) as start_partition,\
          patch('slapos.cli.boot.launchSupervisord', return_value=None),\
          patch('slapos.cli.boot._ping_hostname', return_value=1) as _ping_hostname:
        app.run(('node', 'boot'))

      # boot command runs as root
      check_root_user.assert_called_once()
      # Computer partition was started during boot
      start_partition.assert_called_once()
      # it waits for interface to have an IPv6 address
      ifaddresses.assert_called_once_with('interface_name_from_config')
      # then ping master hostname to wait for connectivity
      _ping_hostname.assert_called_once_with('slap.vifib.com')
      # then format and bang
      SlapOSApp().run.assert_any_call(['node', 'format', '--now', '--local', '--verbose'])
      SlapOSApp().run.assert_any_call(['node', 'bang', '-m', 'Reboot'])

      # timestamp files have been removed
      self.assertFalse(os.path.exists(timestamp),
              timestamp)

  def test_boot_failure(self):
    # In this test, the network interfaces will not have
    # IP address at the beginning, the global IPv6 address only appears later,
    # then format and bang commands will fail two time each.
    # `slapos node boot` command retries on failures.

    app = slapos.cli.entry.SlapOSApp()
    net1 = {socket.AF_INET: ({'addr': '127.0.0.1'},),}
    net2 = {socket.AF_INET: ({'addr': '127.0.0.1'},), socket.AF_INET6: ({'addr': 'fe80::1'},),}
    net3 = {socket.AF_INET: ({'addr': '127.0.0.1'},), socket.AF_INET6: ({'addr': 'fe80::1'}, {'addr': '2000::1'},),}
    with patch('slapos.cli.boot.check_root_user', return_value=True) as check_root_user,\
        patch('slapos.cli.boot.sleep') as sleep,\
        patch('slapos.cli.boot.netifaces.ifaddresses',
              side_effect=[net1, net2, net3]),\
        patch('slapos.cli.boot._ping_hostname', return_value=0),\
        patch('slapos.cli.boot._startComputerPartitionList', return_value=None) as start_partition,\
        patch('slapos.cli.format.check_root_user', return_value=True),\
        patch('slapos.cli.format.logging.FileHandler', return_value=logging.NullHandler()),\
        patch('slapos.cli.bang.check_root_user', return_value=True),\
        patch('slapos.cli.format.do_format', side_effect=[Exception, Exception, None]) as do_format,\
        patch('slapos.cli.bang.do_bang', side_effect=[Exception, Exception, None]) as do_bang:

      app.run(('node', 'boot'))

    check_root_user.assert_called_once()
    start_partition.assert_called_once()

    self.assertEqual(do_format.call_count, 3)
    self.assertEqual(do_bang.call_count, 3)

    # between retries of ping, we sleep 5 seconds
    # between retries of bang, we sleep 15 seconds
    self.assertEqual(sleep.mock_calls, [mock.call(5)]*2 + [mock.call(15)]*4)

    # we have only one logger on the console
    from slapos.cli import coloredlogs
    self.assertEqual(
        len([
            h for h in logging.getLogger('').handlers
            if isinstance(h, coloredlogs.ColoredStreamHandler)
        ]), 1)


class TestCliNode(CliMixin):

  def test_node_software(self):
    """slapos node software command
    """
    app = slapos.cli.entry.SlapOSApp()

    software_release = mock.MagicMock()
    software_release.getState = mock.Mock(return_value='available')
    software_release.getURI = mock.Mock(return_value='http://example.org/software.cfg')
    software_release.building = mock.Mock()

    computer = mock.MagicMock()
    computer.getSoftwareReleaseList = mock.Mock(return_value=[software_release])

    software = mock.MagicMock()
    from slapos.grid.slapgrid import Slapgrid
    from slapos.slap.slap import slap
    with patch('slapos.cli.slapgrid.check_root_user', return_value=True) as checked_root_user, \
         patch('slapos.cli.slapgrid.setRunning') as write_pid_file, \
         patch.object(Slapgrid, 'checkEnvironmentAndCreateStructure') as checkEnvironmentAndCreateStructure, \
         patch.object(slap, 'registerComputer', return_value=computer) as registerComputer, \
         patch('slapos.grid.slapgrid.Software', return_value=software) as Software, \
         patch('slapos.grid.slapgrid.open') as _open:

      app.run(('node', 'software'))

      checked_root_user.assert_called_once()
      write_pid_file.assert_called_once_with(
          logger=mock.ANY,
          pidfile='/opt/slapos/slapgrid-sr.pid')
      checkEnvironmentAndCreateStructure.assert_called_once()
      registerComputer.assert_called_once()

      software_constructor_call, = Software.call_args_list
      self.assertEqual('http://example.org/software.cfg', software_constructor_call[1]['url'])
      # by default software are not built in debug mode
      self.assertFalse(software_constructor_call[1]['buildout_debug'])

      software.install.assert_called_once()

  def test_node_instance(self):
    """slapos node instance command
    """
    app = slapos.cli.entry.SlapOSApp()

    from slapos.grid.slapgrid import Slapgrid
    with patch('slapos.cli.slapgrid.check_root_user', return_value=True) as checked_root_user, \
         patch('slapos.cli.slapgrid.setRunning') as write_pid_file, \
         patch.object(Slapgrid, 'processComputerPartitionList') as processComputerPartitionList:

      app.run(('node', 'instance'))

      checked_root_user.assert_called_once()
      write_pid_file.assert_called_once_with(
          logger=mock.ANY,
          pidfile='/opt/slapos/slapgrid-cp.pid')
      processComputerPartitionList.assert_called_once()

  def test_node_prune(self):
    """slapos node prune command
    """
    app = slapos.cli.entry.SlapOSApp()

    with patch('slapos.cli.prune.check_root_user', return_value=True) as checked_root_user, \
         patch('slapos.cli.prune.setRunning') as write_pid_file, \
         patch('slapos.cli.prune.merged_options', return_value={
            'shared_part_list': 'something',
            'root_check': 'true',
            'pidfile_software': 'pidfile_software.pid',
         }), \
         patch('slapos.cli.prune.do_prune') as do_prune:

      app.run(('node', 'prune'))

      checked_root_user.assert_called_once()
      write_pid_file.assert_called_once_with(
          logger=mock.ANY,
          pidfile='pidfile_software.pid')
      do_prune.assert_called_once()


class TestCliList(CliMixin):
  def test_list(self):
    """
    Test "slapos service list" command output.
    """
    return_value = {
      'instance1': slapos.slap.SoftwareInstance(_title='instance1', _software_release_url='SR1'),
      'instance2': slapos.slap.SoftwareInstance(_title='instance2', _software_release_url='SR2'),
    }
    with patch.object(slapos.slap.slap, 'getOpenOrderDict', return_value=return_value) as _:
      slapos.cli.list.do_list(self.logger, self.conf, self.local)
    if six.PY3:
      self.logger.info.assert_called_once_with('{\n  "instance1": "SR1",\n  "instance2": "SR2"\n}')

  def test_emptyList(self):
    with patch.object(slapos.slap.slap, 'getOpenOrderDict', return_value={}) as _:
      slapos.cli.list.do_list(self.logger, self.conf, self.local)

    self.logger.info.assert_called_once_with('{}')

@patch.object(slapos.slap.slap, 'registerOpenOrder', return_value=slapos.slap.OpenOrder())
class TestCliInfo(CliMixin):

  def test_info_xml_serialisation(self, _):
    self._test_info_output(
      slapos.slap.SoftwareInstance(
        _software_release_url='SR1',
        _requested_state='mystate',
        _connection_dict=
        '<?xml version=\'1.0\' encoding=\'utf-8\'?>\n<instance>\n  <parameter id="myconnectionparameter">value1</parameter>\n</instance>\n',
        _parameter_dict={'myinstanceparameter': 'value2'}))

  def test_info_json_in_serialisation(self, _):
    self._test_info_output(
      slapos.slap.SoftwareInstance(
        _software_release_url='SR1',
        _requested_state='mystate',
        _connection_dict=
        '<?xml version=\'1.0\' encoding=\'utf-8\'?>\n<instance>\n  <parameter id="_">{"myconnectionparameter": "value1"}</parameter>\n</instance>\n',
        _parameter_dict={'myinstanceparameter': 'value2'}))

  def _test_info_output(self, instance):
    """
    Test "slapos service info" command output.
    """
    setattr(self.conf, 'reference', 'instance1')
    with patch.object(slapos.slap.OpenOrder, 'getInformation',
                      return_value=instance):
      slapos.cli.info.do_info(self.logger, self.conf, self.local)

    if six.PY3:
      self.logger.info.assert_called_once_with(
        textwrap.dedent('''\
        {
          "software-url": "SR1",
          "instance-state": "mystate",
          "instance-parameters": {
            "myinstanceparameter": "value2"
          },
          "connection-parameters": {
            "myconnectionparameter": "value1"
          }
        }'''))

  def test_unknownReference(self, _):
    """
    Test "slapos service info" command output in case reference
    of service is not known.
    """
    setattr(self.conf, 'reference', 'instance1')
    with patch.object(slapos.slap.OpenOrder, 'getInformation', side_effect=raiseNotFoundError):
      slapos.cli.info.do_info(self.logger, self.conf, self.local)

    self.logger.warning.assert_called_once_with('Instance %s does not exist.', self.conf.reference)

class TestCliComputerList(CliMixin):
  def test_computer_list(self):
    """
    Test "slapos computer list" command output.
    """
    return_value = {
      'computer1': slapos.slap.hateoas.TempDocument(title='computer1', _reference='COMP-1'),
      'computer2': slapos.slap.hateoas.TempDocument(title='computer2', _reference='COMP-0'),
    }
    with patch.object(slapos.slap.slap, 'getComputerDict', return_value=return_value) as _:
      slapos.cli.computer_list.do_list(self.logger, None, self.local)

    self.logger.info.assert_any_call('%s %s', 'COMP-1', 'computer1')
    self.logger.info.assert_any_call('%s %s', 'COMP-0', 'computer2')

  def test_computer_emptyList(self):
    with patch.object(slapos.slap.slap, 'getComputerDict', return_value={}) as _:
      slapos.cli.computer_list.do_list(self.logger, None, self.local)

    self.logger.info.assert_called_once_with('No existing computer.')

@patch.object(slapos.slap.slap, 'registerComputer', return_value=slapos.slap.Computer("COMP-1"))
class TestComputerCliInfo(CliMixin):
  def test_computer_info(self, _):
    """
    Test "slapos computer info" command output.
    """
    setattr(self.conf, 'reference', 'COMP-1')
    computer = slapos.slap.Computer("COMP-1")
    computer._reference = "COMP-1"
    computer._title = "computer1"
    with patch.object(slapos.slap.Computer, 'getInformation', return_value=computer):
      slapos.cli.computer_info.do_info(self.logger, self.conf, self.local)

    self.logger.info.assert_any_call('Computer Reference: %s', computer._reference)
    self.logger.info.assert_any_call('Computer Title    : %s', computer._title)

  def test_computer_unknownReference(self, _):
    """
    Test "slapos computer info" command output in case reference
    of computer is not known.
    """
    setattr(self.conf, 'reference', 'COMP-0')
    with patch.object(slapos.slap.Computer, 'getInformation', side_effect=raiseNotFoundError):
      slapos.cli.computer_info.do_info(self.logger, self.conf, self.local)

    self.logger.warning.assert_called_once_with('Computer %s does not exist.', self.conf.reference)

@patch.object(slapos.slap.slap, 'registerToken', return_value=slapos.slap.Token())
class TestComputerCliToken(CliMixin):
  def test_computer_token(self, _):
    """
    Test "slapos computer token" command output.
    """
    token = "1234567-90"
    with patch.object(slapos.slap.Token, 'request', return_value=token):
      slapos.cli.computer_token.do_token(self.logger, self.conf, self.local)

    self.logger.info.assert_any_call('Computer token: %s', "1234567-90")

@patch.object(supervisor.supervisorctl, 'main')
class TestCliSupervisorctl(CliMixin):
  def test_allow_supervisord_launch(self, _):
    """
    Test that "slapos node supervisorctl" tries to launch supervisord
    """
    instance_root = '/foo/bar'
    with patch.object(slapos.grid.svcbackend, 'launchSupervisord') as launchSupervisord:
      slapos.cli.supervisorctl.do_supervisorctl(self.logger, instance_root, ['status'], False)
      launchSupervisord.assert_any_call(instance_root=instance_root, logger=self.logger)

  def test_forbid_supervisord_launch(self, _):
    """
    Test that "slapos node supervisorctl" does not try to launch supervisord
    """
    instance_root = '/foo/bar'
    with patch.object(slapos.grid.svcbackend, 'launchSupervisord') as launchSupervisord:
      slapos.cli.supervisorctl.do_supervisorctl(self.logger, instance_root, ['status'], True)
      self.assertFalse(launchSupervisord.called)


class TestCliConsole(unittest.TestCase):

  default_script = """\
print(request('software_release', 'instance').getInstanceParameterDict()['parameter_name'])
"""

  @contextmanager
  def _test_console(self):
    cp = slapos.slap.ComputerPartition('computer_%s' % self.id(), 'partition_%s' % self.id())
    cp._parameter_dict = {'parameter_name': 'parameter_value'}
    with patch.object(slapos.slap.OpenOrder, 'request',
                      return_value = cp) as mock_request, \
         patch.object(sys, 'stdout', StringIO()) as app_stdout, \
         tempfile.NamedTemporaryFile() as config_file:
      config_file.write(b'[slapos]\nmaster_url=null\n')
      config_file.flush()
      yield slapos.cli.entry.SlapOSApp(), \
          config_file.name, \
          mock_request, \
          app_stdout


  def test_console_interactive(self):
    with self._test_console() as (app, config_file, mock_request, stdout), \
         patch.object(sys, 'stdin', StringIO(self.default_script)):
      app.run(('console', '--cfg', config_file))
      self.assertIn('parameter_value', stdout.getvalue())
      mock_request.assert_called_once_with(
          'software_release', 'instance')

  def test_console_interactive_no__file__(self):
    script = '''if 1:
      try:
        print('FAIL __file__ is set to', __file__)
      except NameError:
        print('OK __file__ is not set')

      '''
    with self._test_console() as (app, config_file, _, stdout), \
         patch.object(sys, 'stdin', StringIO(script)):
      app.run(('console', '--cfg', config_file))
      self.assertIn('OK __file__ is not set', stdout.getvalue())

  def test_console_script(self):
    with self._test_console() as (app, config_file, mock_request, stdout), \
         tempfile.NamedTemporaryFile('w') as script:
      script.write(self.default_script)
      script.flush()
      app.run(('console', '--cfg', config_file, script.name))
      self.assertIn('parameter_value', stdout.getvalue())
      mock_request.assert_called_once_with(
          'software_release', 'instance')

  def test_console_script__file__(self):
    with self._test_console() as (app, config_file, _, stdout),\
      tempfile.NamedTemporaryFile('w') as script:
      script.write('''if 1:
        if __file__ == %r:
          print('OK __file__ is set to script')
        else:
          print('FAIL __file__ is set to', __file__)
      ''' % script.name)
      script.flush()
      app.run(('console', '--cfg', config_file, script.name))
      self.assertIn('OK __file__ is set to script', stdout.getvalue())


class TestCliComplete(CliMixin):
  def test_complete_bash(self):
    with patch.object(sys, 'stdout', StringIO()) as app_stdout:
      self.assertEqual(slapos.cli.entry.SlapOSApp().run(['complete']), 0)
    self.assertIn('COMPREPLY', app_stdout.getvalue())

  def test_complete_fish(self):
    with patch.object(sys, 'stdout', StringIO()) as app_stdout:
      self.assertEqual(slapos.cli.entry.SlapOSApp().run(['complete', '--shell=fish']), 0)
    self.assertIn('__fish_seen_subcommand_from', app_stdout.getvalue())


class TestCliRequest(CliMixin):
  def test_parse_option_dict(self):
    parse_option_dict = slapos.cli.request.parse_option_dict

    self.assertEqual(parse_option_dict(['foo=bar', 'a=b']), {'foo': 'bar', 'a': 'b'})
    # malformed option = assignment
    self.assertRaises(ValueError, parse_option_dict, ['a'])
    # duplicated key
    self.assertRaises(ValueError, parse_option_dict, ['a=b', 'a=c'])
    # corner cases
    self.assertEqual(parse_option_dict(['a=a=b']), {'a': 'a=b'})
    self.assertEqual(parse_option_dict(['a=a\nb']), {'a': 'a\nb'})
    self.assertEqual(parse_option_dict([]), {})

  def test_request(self):
    self.conf.reference = 'instance reference'
    self.conf.software_url = 'software URL'
    self.conf.parameters = {'key': 'value'}
    self.conf.parameters_file = None
    self.conf.node = {'computer_guid': 'COMP-1234'}
    self.conf.type = None
    self.conf.state = None
    self.conf.slave = False

    with patch.object(
        slapos.slap.slap,
        'registerOpenOrder',
        return_value=mock.create_autospec(slapos.slap.OpenOrder)) as registerOpenOrder:
      slapos.cli.request.do_request(self.logger, self.conf, self.local)

    registerOpenOrder().request.assert_called_once_with(
        software_release='software URL',
        partition_reference='instance reference',
        partition_parameter_kw={'key': 'value'},
        software_type=None,
        filter_kw={'computer_guid': 'COMP-1234'},
        state=None,
        shared=False,
    )

    self.logger.info.assert_any_call(
        'Requesting %s as instance of %s...',
        'instance reference',
        'software URL',
    )

  def test_request_json_in_xml_published_parameters(self):
    tmpdir = tempfile.mkdtemp()
    self.addCleanup(shutil.rmtree, tmpdir)
    with open(os.path.join(tmpdir, 'software.cfg.json'), 'w') as f:
      json.dump(
          {
              "name": "Test Software",
              "description": "Dummy software for Test",
              "serialisation": "json-in-xml",
              "software-type": {
                  DEFAULT_SOFTWARE_TYPE: {
                      "title": "Default",
                      "description": "Default type",
                      "request": "instance-default-input-schema.json",
                      "response": "instance-default-output-schema.json",
                      "index": 0
                  },
              }
          }, f)

    self.conf.reference = 'instance reference'
    self.conf.software_url = os.path.join(tmpdir, 'software.cfg')
    self.conf.parameters = {'key': 'value'}
    self.conf.parameters_file = None
    self.conf.node = {'computer_guid': 'COMP-1234'}
    self.conf.type = None
    self.conf.state = None
    self.conf.slave = False

    cp = slapos.slap.ComputerPartition(
        'computer_%s' % self.id(),
        'partition_%s' % self.id())
    cp._requested_state = 'started'
    cp._connection_dict = {'_': json.dumps({'foo': 'bar'})}

    with patch.object(
        slapos.slap.slap,
        'registerOpenOrder',
        return_value=mock.create_autospec(slapos.slap.OpenOrder)) as registerOpenOrder:
      registerOpenOrder().request.return_value = cp
      slapos.cli.request.do_request(self.logger, self.conf, self.local)

    registerOpenOrder().request.assert_called_once()

    self.assertEqual(self.logger.info.mock_calls, [
        mock.call('Requesting %s as instance of %s...', self.conf.reference,
                  self.conf.software_url),
        mock.call('Instance requested.\nState is : %s.', 'started'),
        mock.call('Connection parameters of instance are:'),
        mock.call("{'foo': 'bar'}"),
        mock.call('You can rerun the command to get up-to-date information.'),
    ])


class TestCliRequestParameterFile(CliMixin):
  """Request with --parameter-file, with a .json file.
  """
  expected_partition_parameter_kw = {'foo': ['bar']}

  def _makeParameterFile(self):
    f = tempfile.NamedTemporaryFile(
        suffix=self.parameter_file_suffix,
        mode='w', delete=False,
    )
    self.addCleanup(os.unlink, f.name)
    f.write(textwrap.dedent(self.parameter_file_content))
    f.flush()
    return f.name

  def _request_parameters_file_setup(self):
    self.conf.reference = 'instance reference'
    self.conf.software_url = 'software URL'
    self.conf.parameters =  None
    f = open(self._makeParameterFile())
    self.addCleanup(f.close)
    self.conf.parameters_file = f
    self.conf.node = {'computer_guid': 'COMP-1234'}
    self.conf.type = None
    self.conf.state = None
    self.conf.slave = False


class TestCliRequestParameterFileUndefinedSerialization(TestCliRequestParameterFile):
  """Request with --parameter-file, without defining serialization type.
  """
  parameter_file_suffix = ''
  parameter_file_content = '''\
  {
    "foo": ["bar"]
  }
  '''

  def test_request_parameters_file(self):
    self._request_parameters_file_setup()
    self.assertRaises(
        UndefinedSerializationError,
        slapos.cli.request.do_request,
        self.logger,
        self.conf,
        self.local,
    )


class TestCliRequestParametersFileJson(TestCliRequestParameterFile):
  """Request with --parameter-file, with a .json file.
  """
  parameter_file_suffix = '.json'
  parameter_file_content = '''\
  {
    "foo": ["bar"]
  }
  '''
  serialization = 'xml'

  def test_request_parameters_file(self):
    self._request_parameters_file_setup()
    with mock.patch(
        'slapos.cli.request.SoftwareReleaseSchema.getSerialisation',
        return_value=self.serialization):
      with patch.object(
          slapos.slap.slap,
          'registerOpenOrder',
          return_value=mock.create_autospec(slapos.slap.OpenOrder)) as registerOpenOrder:
        slapos.cli.request.do_request(self.logger, self.conf, self.local)

      registerOpenOrder().request.assert_called_once_with(
          software_release='software URL',
          partition_reference='instance reference',
          partition_parameter_kw=self.expected_partition_parameter_kw,
          software_type=None,
          filter_kw={'computer_guid': 'COMP-1234'},
          state=None,
          shared=False,
      )
      self.logger.info.assert_any_call(
          'Requesting %s as instance of %s...',
          'instance reference',
          'software URL',
      )


class TestCliRequestParametersFileJsonJsonInXMLSerialisation(
    TestCliRequestParametersFileJson):
  """Request with --parameter-file, with a .json file and a software using
  json-in-xml for serialisation. In that case, the parameters are automatically
  serialised with {'_': json.dumps(params)}
  """
  expected_partition_parameter_kw = {"_": "{\"foo\": [\"bar\"]}"}
  serialization = 'json-in-xml'


class TestCliRequestParametersFileJsonJsonInXMLSerialisationAlreadySerialised(
    TestCliRequestParametersFileJsonJsonInXMLSerialisation):
  """Request with --parameter-file, with a .json file and a software using
  json-in-xml for serialisation and parameters already serialised with
  {'_': json.dumps(params)}. In that case, parameters are not serialized one
  more time.
  """
  expected_partition_parameter_kw = {"_": "{\"foo\": [\"bar\"]}"}
  parameter_file_content = r'''
    {"_": "{\"foo\": [\"bar\"]}"}
  '''


class TestCliRequestParametersFileYaml(TestCliRequestParametersFileJson):
  """Request with --parameter-file, with a .yaml file. This behaves like json.
  """
  parameter_file_content = '''\
    foo:
    - bar
  '''
  parameter_file_suffix = '.yaml'


class TestCliRequestParametersFileXml(TestCliRequestParametersFileJson):
  """Request with --parameter-file, with a .xml file
  """
  expected_partition_parameter_kw = {'foo': 'bar'}
  parameter_file_content = '''\
    <?xml version="1.0" encoding="utf-8"?>
    <instance>
        <parameter id="foo">bar</parameter>
    </instance>
  '''
  parameter_file_suffix = '.xml'



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2010-2014 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import argparse
import codecs
import collections
import locale
import logging
import platform
import sys

import os
# hack to avoid a bug in cmd2: https://bitbucket.org/catherinedevlin/cmd2/issue/1/silent-editor-check
# must be done before importing cliff
os.environ.setdefault('EDITOR', 'vi')

from cliff.app import App
from cliff.commandmanager import CommandManager, LOG
from requests.packages import urllib3

import slapos.version
import slapos.cli.complete

# silence messages like 'Starting connection' that are logged with INFO
urllib3_logger = logging.getLogger('requests.packages.urllib3')
urllib3_logger.setLevel(logging.WARNING)


class SlapOSCommandManager(CommandManager):

    def find_command(self, argv):
        """Given an argument list, find a command and
        return the processor and any remaining arguments.
        """
        # a little cheating, 'slapos node' is not documented by the help command
        if argv == ['node']:
            argv = ['node', 'status']

        search_args = argv[:]
        name = ''
        while search_args:
            if search_args[0].startswith('-'):
                LOG.critical('slapos: invalid option %r' % search_args[0])
                sys.exit(5)
            next_val = search_args.pop(0)
            name = '%s %s' % (name, next_val) if name else next_val
            if name in self.commands:
                cmd_ep = self.commands[name]
                cmd_factory = cmd_ep.load()
                return (cmd_factory, name, search_args)
        else:
            LOG.critical('slapos: the command %r does not exist or is not yet implemented.\n'
                         '\n'
                         'Available commands: %s\n\n'
                         'Please find documentation and forum at http://community.slapos.org\n'
                         'Please also make sure that the SlapOS Node package is up to date.',
                         ' '.join(argv),
                         ', '.join(sorted(repr(c) for c in self.commands)))
            sys.exit(5)


class SlapOSHelpAction(argparse.Action):
    """
    Adapted from cliff.help.HelpAction, this class detects
    and outputs command groups, via the .command_group attribute
    of the Command class. Must be a class attribute in case the class
    cannot be instantiated ('Could not load' message).
    """
    def __call__(self, parser, namespace, values, option_string=None):
        app = self.default
        parser.print_help(app.stdout)
        command_manager = app.command_manager
        groups = collections.defaultdict(list)
        for name, ep in sorted(command_manager):
            command_group, help_line = self._help_line(ep, name)
            groups[command_group].append(help_line)

        for group in sorted(groups):
            app.stdout.write('\n%s commands:\n' % group)
            for line in sorted(groups[group]):
                app.stdout.write(line)
        sys.exit(0)

    def _help_line(self, ep, name):
        try:
            factory = ep.load()
        except Exception as err:
            return 'Could not load %r\n' % ep
        try:
            cmd = factory(self, None)
        except Exception as err:
            return 'Could not instantiate %r: %s\n' % (ep, err)
        one_liner = cmd.get_description().split('\n')[0]
        group = getattr(factory, 'command_group', 'other')
        return group, '  %-13s  %s\n' % (name, one_liner)


_logging_configured = []

class SlapOSApp(App):

    #
    # self.options.verbose_level:
    #    -q  -> 0 (WARNING)
    #        -> 1 (INFO)
    #    -v  -> 2 (DEBUG)
    #    -vv -> 3 (...)
    #    etc.
    #

    log = logging.getLogger('slapos')

    def __init__(self):
        super(SlapOSApp, self).__init__(
            description='SlapOS client %s' % slapos.version.version,
            version='%s (Python %s)' % (slapos.version.version, platform.python_version()),
            command_manager=SlapOSCommandManager('slapos.cli'),
        )
        # replace default complete command
        self.command_manager.add_command('complete', slapos.cli.complete.CompleteCommand)


    def _set_streams(self, stdin, stdout, stderr):
        try:
            # SlapOS: might fail in some systems
            locale.setlocale(locale.LC_ALL, '')
        except locale.Error:
            pass

        if sys.version_info[:2] == (2, 6):
            # Configure the input and output streams. If a stream is
            # provided, it must be configured correctly by the
            # caller. If not, make sure the versions of the standard
            # streams used by default are wrapped with encodings. This
            # works around a problem with Python 2.6 fixed in 2.7 and
            # later (http://hg.python.org/cpython/rev/e60ef17561dc/).
            lang, encoding = locale.getdefaultlocale()
            encoding = getattr(sys.stdout, 'encoding', None) or encoding
            self.stdin = stdin or codecs.getreader(encoding)(sys.stdin)
            self.stdout = stdout or codecs.getwriter(encoding)(sys.stdout)
            self.stderr = stderr or codecs.getwriter(encoding)(sys.stderr)
        else:
            self.stdin = stdin or sys.stdin
            self.stdout = stdout or sys.stdout
            self.stderr = stderr or sys.stderr

    def build_option_parser(self, *args, **kw):
        kw.setdefault('argparse_kwargs', {})
        kw['argparse_kwargs']['conflict_handler'] = 'resolve'
        parser = super(SlapOSApp, self).build_option_parser(*args, **kw)

        # add two aliases for --log-file (for compatibility with old commands)
        parser.add_argument(
            '--log-file', '--logfile', '--log_file',
            action='store',
            default=None,
            help='Specify a file to log output (default: console only)',
        )
        parser.add_argument(
            '--log-color',
            action='store_true',
            help='Colored log output in console (stripped if redirected)',
            default=True,
        )
        parser.add_argument(
            '--log-time',
            action='store_false',
            default=True,
            help='Include timestamp in console log',
        )
        parser.add_argument(
            '-h', '--help',
            action=SlapOSHelpAction,
            nargs=0,
            default=self,  # tricky
            help="show this help message and exit",
        )
        return parser

    def initialize_app(self, argv):
        if self.options.verbose_level > 2:
            self.log.debug('initialize_app')

    def prepare_to_run_command(self, cmd):
        if self.options.verbose_level > 2:
            self.log.debug('prepare_to_run_command %s', cmd.__class__.__name__)

    def clean_up(self, cmd, result, err):
        if self.options.verbose_level > 2:
            self.log.debug('clean_up %s', cmd.__class__.__name__)

    @property
    def CONSOLE_MESSAGE_FORMAT(self):
        fmt = []
        if self.options.log_time and not self.options.log_color:
            fmt.append('[%(asctime)s]')
        if not self.options.log_color:
            fmt.append('%(levelname)s')
        fmt.append('%(message)s')
        return ' '.join(fmt)

    @property
    def LOG_FILE_MESSAGE_FORMAT(self):
        return '[%(asctime)s] %(levelname)-8s %(message)s'

    def configure_logging(self):
        """Create logging handlers for any log output.
        """
        if _logging_configured:
            # This function is called again if "slapos node boot" failed.
            # Don't add a handler again, otherwise the output becomes double.
            return

        root_logger = logging.getLogger('')
        root_logger.setLevel(logging.DEBUG)

        # Set up logging to a file
        if self.options.log_file:
            file_handler = logging.FileHandler(
                filename=self.options.log_file,
            )
            formatter = logging.Formatter(self.LOG_FILE_MESSAGE_FORMAT)
            file_handler.setFormatter(formatter)
            root_logger.addHandler(file_handler)

        # Always send higher-level messages to the console via stderr
        if self.options.log_color:
            from slapos.cli import coloredlogs
            console = coloredlogs.ColoredStreamHandler(show_name=True,     # logger name (slapos) and PID
                                                       show_severity=True,
                                                       show_timestamps=self.options.log_time,
                                                       show_hostname=False)
        else:
            console = logging.StreamHandler(self.stderr)
        console_level = {0: logging.WARNING,
                         1: logging.INFO,
                         2: logging.DEBUG,
                         }.get(self.options.verbose_level, logging.DEBUG)
        console.setLevel(console_level)
        formatter = logging.Formatter(self.CONSOLE_MESSAGE_FORMAT)
        console.setFormatter(formatter)
        root_logger.addHandler(console)
        _logging_configured.append(True)


    def run(self, argv):
        # same as cliff.app.App.run except that it won't re-raise
        # a logged exception, and doesn't use --debug
        self.options, remainder = self.parser.parse_known_args(argv)
        self.configure_logging()
        self.interactive_mode = not remainder
        try:
            self.initialize_app(remainder)
        except Exception as err:
            self.log.exception(err)
            return 1
        if self.interactive_mode:
            result = self.interact()
        else:
            result = self.run_subcommand(remainder)
        return result

    def run_subcommand(self, argv):
        # same as cliff.app.App.run_subcommand except that it won't re-raise
        # a logged exception, and doesn't use --debug
        subcommand = self.command_manager.find_command(argv)
        cmd_factory, cmd_name, sub_argv = subcommand
        cmd = cmd_factory(self, self.options)
        err = None
        result = 1
        try:
            self.prepare_to_run_command(cmd)
            full_name = (cmd_name
                         if self.interactive_mode
                         else ' '.join([self.NAME, cmd_name])
                         )
            cmd_parser = cmd.get_parser(full_name)
            parsed_args = cmd_parser.parse_args(sub_argv)
            result = cmd.run(parsed_args)
        except Exception as err:
            self.log.exception(err)
            try:
                self.clean_up(cmd, result, err)
            except Exception as err2:
                self.log.exception(err2)
        else:
            try:
                self.clean_up(cmd, result, None)
            except Exception as err3:
                self.log.exception(err3)
        return result


def main(argv=sys.argv[1:]):
    app = SlapOSApp()
    if not argv:
        argv = ['-h']
    return app.run(argv)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



# coding: utf-8
import importlib 
import re
import six

from zope.interface import declarations

config_option = "manager_list"


def load_manager(name):
    """Load a manager from local files if it exists."""
    if re.match(r'[a-zA-Z_]', name) is None:
      raise ValueError("Manager name \"{!s}\" is not allowed! Must contain only letters and \"_\"".
        format(name))
    manager_module_name = "slapos.manager.{}".format(name)

    from slapos.manager import interface
    manager_module = importlib.import_module(manager_module_name)
    if not hasattr(manager_module, "Manager"):
      raise AttributeError("Manager class in {} has to be called \"Manager\"".format(
        manager_module_name))
    if not interface.IManager.implementedBy(manager_module.Manager):
      raise RuntimeError("Manager class in {} must zope.interface.implements \"IManager\"".format(
        manager_module_name))

    return manager_module.Manager


def from_config(config):
    """Return list of instances of managers allowed from the config."""
    if config is None:
      return []

    name_list = config.get(config_option, "")
    if isinstance(name_list, six.string_types):
      name_list = name_list.replace(",", " ").split()

    return [load_manager(name)(config) for name in name_list]


# -*- coding: utf-8 -*-
# vim: set et sts=2:
##############################################################################
#
# Copyright (c) 2019 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import json
import six
from six.moves.urllib import parse
from uritemplate import expand
import os
import logging

from ..util import _addIpv6Brackets
from .exception import ResourceNotReady, NotFoundError, \
          AuthenticationError, ConnectionError

import requests
# silence messages like 'Starting connection' that are logged with INFO
urllib3_logger = logging.getLogger('requests.packages.urllib3')
urllib3_logger.setLevel(logging.WARNING)

from cachecontrol import CacheControl
from cachecontrol.caches.file_cache import FileCache

# XXX fallback_logger to be deprecated together with the old CLI entry points.
fallback_logger = logging.getLogger(__name__)
fallback_handler = logging.StreamHandler()
fallback_logger.setLevel(logging.INFO)
fallback_logger.addHandler(fallback_handler)

ALLOWED_JIO_FIELD_LIST = [
  "StringField",
  "EmailField",
  "IntegerField",
  "FloatField",
  "TextAreaField"]

class TempDocument(object):
  def __init__(self, **kw):
    """
    Makes easy initialisation of class parameters
    """
    self.__dict__.update(kw)

class ConnectionHelper:
  def __init__(self, master_url, key_file=None,
               cert_file=None, master_ca_file=None, timeout=None):
    master_url = _addIpv6Brackets(master_url)
    if master_url.endswith('/'):
        self.slapgrid_uri = master_url
    else:
        # add a slash or the last path segment will be ignored by urljoin
        self.slapgrid_uri = master_url + '/'
    self.key_file = key_file
    self.cert_file = cert_file
    self.master_ca_file = master_ca_file
    self.timeout = timeout

    # self.session will handle requests using HTTP Cache Control rules.
    self.uncached_session = requests.Session()
    self.session = CacheControl(self.uncached_session,
      cache=FileCache(os.path.expanduser("~/.slapos_cached_get")))

  def do_request(self, method, path, params=None, data=None, headers=None):
    url = parse.urljoin(self.slapgrid_uri, path)
    if headers is None:
      headers = {}
    headers.setdefault('Accept', '*/*')
    if path.startswith('/'):
      path = path[1:]
#      raise ValueError('method path should be relative: %s' % path)

    try:
      if url.startswith('https'):
        cert = (self.cert_file, self.key_file)
      else:
        cert = None

      # XXX TODO: handle host cert verify

      # Old behavior was to pass empty parameters as "None" value.
      # Behavior kept for compatibility with old slapproxies (< v1.3.3).
      # Can be removed when old slapproxies are no longer in use.
      if data:
        for k, v in six.iteritems(data):
          if v is None:
            data[k] = 'None'

      req = method(url=url,
                   params=params,
                   cert=cert,
                   verify=False,
                   data=data,
                   headers=headers,
                   timeout=self.timeout)
      try:
        req.raise_for_status()
      except TypeError:
        # In Py3, a comparison between NoneType and int can occur if req has no
        # status_code (= None).
        pass

    except (requests.Timeout, requests.ConnectionError) as exc:
      raise ConnectionError("Couldn't connect to the server. Please "
                            "double check given master-url argument, and make sure that IPv6 is "
                            "enabled on your machine and that the server is available. The "
                            "original error was:\n%s" % exc)
    except requests.HTTPError as exc:
      if exc.response.status_code == requests.status_codes.codes.not_found:
        msg = url
        if params:
            msg += ' - %s' % params
        raise NotFoundError(msg)
      elif exc.response.status_code == requests.status_codes.codes.request_timeout:
        # this is explicitly returned by SlapOS master, and does not really mean timeout
        raise ResourceNotReady(path)
        # XXX TODO test request timeout and resource not found
      else:
        # we don't know how or don't want to handle these (including Unauthorized)
        req.raise_for_status()
    except requests.exceptions.SSLError as exc:
      raise AuthenticationError("%s\nCouldn't authenticate computer. Please "
                                "check that certificate and key exist and are valid." % exc)

#    XXX TODO parse server messages for client configure and node register
#    elif response.status != httplib.OK:
#      message = parsed_error_message(response.status,
#                                     response.read(),
#                                     path)
#      raise ServerError(message)

    return req

  def GET(self, path, params=None, headers=None):
    req = self.do_request(self.session.get,
                          path=path,
                          params=params,
                          headers=headers)
    return req.text.encode('utf-8')

  def POST(self, path, params=None, data=None,
           content_type='application/x-www-form-urlencoded'):
    req = self.do_request(requests.post,
                          path=path,
                          params=params,
                          data=data,
                          headers={'Content-type': content_type})
    return req.text.encode('utf-8')

class HateoasNavigator(object):
  """
  Navigator for HATEOAS-style APIs.
  Inspired by
  https://git.erp5.org/gitweb/jio.git/blob/HEAD:/src/jio.storage/erp5storage.js
  """
  def __init__(self, slapgrid_uri,
               key_file=None, cert_file=None,
               master_ca_file=None, timeout=60):
    self.slapos_master_hateoas_uri = slapgrid_uri
    self.key_file = key_file
    self.cert_file = cert_file
    self.master_ca_file = master_ca_file
    self.timeout = timeout

  def GET(self, uri, headers=None):
    connection_helper = ConnectionHelper(
        uri, self.key_file, self.cert_file, self.master_ca_file, self.timeout)
    return connection_helper.GET(uri, headers=headers)

  def hateoasGetLinkFromLinks(self, links, title):
    if type(links) == dict:
      if links.get('title') == title:
        return links['href']
      raise NotFoundError('Action %s not found.' % title)
    for action in links:
      if action.get('title') == title:
        return action['href']
    else:
      raise NotFoundError('Action %s not found.' % title)

  def getRelativeUrlFromUrn(self, urn):
    urn_schema = 'urn:jio:get:'
    try:
      _, url = urn.split(urn_schema)
    except ValueError:
      return
    return str(url)

  def getSiteDocument(self, url, headers=None):
    result = self.GET(url, headers)
    return json.loads(result)

  def getRootDocument(self):
    cached_root_document = getattr(self, 'root_document', None)
    if cached_root_document:
      return cached_root_document
    self.root_document = self.getSiteDocument(
        self.slapos_master_hateoas_uri,
        headers={'Cache-Control': 'no-cache'}
    )
    return self.root_document

  def _getSearchUrl(self):
    root_document = self.getRootDocument()
    return root_document["_links"]['raw_search']['href']

  def _getTraverseUrl(self):
    root_document = self.getRootDocument()
    return root_document["_links"]['traverse']['href']

  def _extractPropertyFromFormDict(self, form_dict):
    """ Reimplemenation in python of extractPropertyFromFormJSON of jio.js """
    form = form_dict["_embedded"]["_view"]
    form_data_dict = {}
    converted_dict = {
      "portal_type" : form_dict["_links"]["type"]["name"]
    }

    
    if "parent" in form_dict["_links"]:
      converted_dict["parent_relative_url"] = \
              "/".join(form_dict["_links"]["parent"]["href"].split("/")[-2:])

    form_data_dict["form_id"] = {
      "key": [form["form_id"]["key"]],
      "default": form["form_id"]["default"]
    }

    for key in form:
      field = form[key]
      if key.startswith("my_"):
        key = key[len("my_"):]
      elif key.startswith("your_"):
        key = key[len("your_"):]
      else:
        continue
      
      if field["type"] in ALLOWED_JIO_FIELD_LIST:
        form_data_dict[key] = {
          "default": field.get("default", None),
          "key": field["key"]
        }
        converted_dict[key] = field.get("default", None)

    return {
      "data" : converted_dict,
      "form_data": form_data_dict
    }

  def jio_allDocs(self, query):
    # by default search_url is a Python script (ERP5Document_getHateoas)
    # which enforces a too low limit of what is to be returned (default = 10)
    # as testnode uses this to introspect created ERP5 instance which usually has
    # at least 18 sub instances this limit is too low thus we need to explicitly
    # increase it. 40 is the de-facto default ERP5 preference for number of objects
    # which are usually to be rendered in alistbox 'view' mode
    if 'limit' not in query:
       query['limit'] = 40
    search_url = self._getSearchUrl()
    getter_link = expand(search_url, query)

    catalog_json = self.GET(getter_link)
    catalog_dict = json.loads(catalog_json)

    # Return the same data structure from jio api:
    return {'data': {
        "rows": catalog_dict["_embedded"]["contents"],
        "total_rows": len(catalog_dict["_embedded"]["contents"])
      }
    }

  def jio_get(self, key):
    traverse_url = self._getTraverseUrl()

    # Hardcoded view, but it should come from a site configuration
    view = "slaposjs_view"
    getter_link = expand(traverse_url, {
        "relative_url": key,
        "view": view
        })

    document_json = self.GET(getter_link)
    document_dict = json.loads(document_json)

    return self._extractPropertyFromFormDict(document_dict)

  def jio_getAttachment(self, key, action, options):
    traverse_url = self._getTraverseUrl()

    if action == "view":
      view = "slaposjs_view"
      getter_link = expand(traverse_url, {
        "relative_url": key,
        "view": view
        })
    elif action == "links":
      raise NotImplementedError("links is not implemented for now")
    elif action.startswith(self.slapos_master_hateoas_uri):
      # This is a url, call it directly
      getter_link = action

    document_json = self.GET(getter_link)
    document_dict = json.loads(document_json)

    return document_dict

  def _getMeUrl(self):
    return self.getRelativeUrlFromUrn(
        self.getRootDocument()['_links']['me']['href'])

  def getMeDocument(self):
    # User can be a Person, Software Instance or Computer
    return self.jio_get(self._getMeUrl())

class SlapHateoasNavigator(HateoasNavigator):

  ###############################################################
  # API for Client as a Person
  ###############################################################
  def _getInstanceTreeList(self, title=None, select_list=["title", "url_string"]):
    query_str = 'portal_type:"Instance Tree" AND validation_state:validated'
    if title is not None:
      query_str = 'portal_type:"Instance Tree" AND validation_state:validated AND title:="%s"' % title

    result = self.jio_allDocs(
      query={"query" : query_str, "select_list": select_list})

    return result['data']['rows']

  def _getComputerList(self, title=None, reference=None, select_list=["title", "reference"]):
    query_str = 'portal_type:"Computer" AND validation_state:validated'
    if title is not None:
      query_str += ' AND title:="%s"' % title

    if reference is not None:
      query_str += ' AND reference:="%s"' % reference

    result = self.jio_allDocs(
      query={"query" : query_str, "select_list": select_list})

    return result['data']['rows']

  def getInstanceTreeInstanceList(self, title):
    select_list = ['uid', 'title', 'relative_url', 'portal_type', 
            'url_string', 'text_content', 'getConnectionXmlAsDict',
            'SoftwareInstance_getNewsDict']
    query_str = 'portal_type:("Software Instance", "Slave Instance") AND validation_state:validated'
    query_str += ' AND default_specialise_title:="%s"' % title
    result = self.jio_allDocs(
      query={"query" : query_str, "select_list": select_list})

    return result['data']['rows']

  def getInstanceTreeDict(self, title=None):
    instance_tree_list = self._getInstanceTreeList(title=title)
    instance_tree_dict = {}
    for instance_tree in instance_tree_list:
      software_instance = TempDocument()
      for key, value in six.iteritems(instance_tree):
        if key in ['_links', 'url_string']:
          continue
        setattr(software_instance, '_%s' % key, value)
      setattr(software_instance, '_software_release_url', instance_tree["url_string"])
      instance_tree_dict[software_instance._title] = software_instance

    return instance_tree_dict

  def getComputerDict(self):
    computer_list = self._getComputerList()
    computer_dict = {}
    for computer_json  in computer_list:
      computer = TempDocument()
      for key, value in six.iteritems(computer_json):
        if key in ['_links']:
          continue
        setattr(computer, '_%s' % key, value)
      computer_dict[computer._title] = computer

    return computer_dict

  def getToken(self):
    root_document = self.getRootDocument()
    hateoas_url = root_document['_links']['self']['href']
    token_json = self.jio_getAttachment(
      "computer_module", hateoas_url +  "/computer_module/Base_getComputerToken", {})

    return token_json["access_token"]

  def getInstanceTreeRootSoftwareInstanceInformation(self, reference):
    instance_tree_list = self._getInstanceTreeList(title=reference,
      select_list=["title", "relative_url"])

    assert len(instance_tree_list) <= 1, \
      "There are more them one Instance Tree for this reference"

    instance_tree_jio_key = None
    for instance_tree_candidate in instance_tree_list:
      if instance_tree_candidate.get('title') == reference:
        instance_tree_jio_key = instance_tree_candidate['relative_url']
        break

    if instance_tree_jio_key is None:
      raise NotFoundError('This document does not exist.')

    return self.jio_get(instance_tree_jio_key)

  def _getComputer(self, reference):
    computer_list = self._getComputerList(reference=reference,
      select_list=["reference", "relative_url"])

    assert len(computer_list) <= 1, \
      "There are more them one Computer for this reference"

    computer_jio_key = None
    for computer_candidate in computer_list:
      if computer_candidate.get("reference") == reference:
        computer_jio_key = computer_candidate['relative_url']
        break

    if computer_jio_key is None:
      raise NotFoundError('This computer does not exist.')

    return self.jio_get(computer_jio_key)

  def getSoftwareInstallationList(self, computer_guid=None,
          select_list=['uid', 'relative_url', 'url_string', 'SoftwareInstallation_getNewsDict']):
    query_str = 'portal_type:"Software Installation" AND validation_state:validated'
    if computer_guid is not None:
      query_str += ' AND default_aggregate_reference:="%s"' % computer_guid

    result = self.jio_allDocs(
      query={"query" : query_str, "select_list": select_list})

    return result['data']['rows']

  def getInstanceNews(self, url):
    return self.jio_get(url)['news']

  def getSoftwareInstallationNews(self, computer_guid=None, url=None):
    for si in self.getSoftwareInstallationList(computer_guid):
      if si["url_string"] == url:
        return si['SoftwareInstallation_getNewsDict']['text']
    return "#error no data found"

  def getComputerNews(self, computer_guid):
    return self._getComputer(reference=computer_guid)['data']['news']['computer']['text']

  def getComputerPartitionNews(self, computer_guid, partition_id):
    try:
      return self._getComputer(reference=computer_guid)['data']['news']['partition'][partition_id]['text']
    except KeyError:
      return "#error no data found"





##############################################################################
#
# Copyright (c) 2010 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import logging
import unittest
import slapos.slap
import slapos.client

class TestClient(unittest.TestCase):
  def setUp(self):
    self.called_software_product = None

    class FakeSoftwareProductCollection(object):
      def __init__(inner_self, *args, **kw_args):
        inner_self.__getattr__ = inner_self.get
      def get(inner_self, software_product):
        self.called_software_product = software_product
        return self.software_product_reference


    self.slap = slapos.slap.slap()
    self.product_collection = FakeSoftwareProductCollection(
        logging.getLogger(), self.slap)


  def test_getSoftwareReleaseFromSoftwareString_softwareProduct(self):
    """
    Test that if given software is a Sofwtare Product (i.e matching
    the magic string), it returns the corresponding value of a call to
    SoftwareProductCollection.
    """
    self.software_product_reference = 'foo'
    software_string = '%s%s' % (
        slapos.client.SOFTWARE_PRODUCT_NAMESPACE,
        self.software_product_reference
    )

    slapos.client._getSoftwareReleaseFromSoftwareString(
        logging.getLogger(), software_string, self.product_collection)

    self.assertEqual(
        self.called_software_product,
        self.software_product_reference
    )

  def test_getSoftwareReleaseFromSoftwareString_softwareProduct_emptySoftwareProduct(self):
    """
    Test that if given software is a Software Product (i.e matching
    the magic string), but this software product is empty, it exits.
    """
    self.software_product_reference = 'foo'
    software_string = '%s%s' % (
        slapos.client.SOFTWARE_PRODUCT_NAMESPACE,
        self.software_product_reference
    )

    def fake_get(software_product):
      raise AttributeError()
    self.product_collection.__getattr__ = fake_get

    self.assertRaises(
        SystemExit,
        slapos.client._getSoftwareReleaseFromSoftwareString,
        logging.getLogger(), software_string, self.product_collection
    )

  def test_getSoftwareReleaseFromSoftwareString_softwareRelease(self):
    """
    Test that if given software is a simple Software Release URL (not matching
    the magic string), it is just returned without modification.
    """
    software_string = 'foo'
    returned_value = slapos.client._getSoftwareReleaseFromSoftwareString(
        logging.getLogger(), software_string, self.product_collection)

    self.assertEqual(
        self.called_software_product,
        None
    )

    self.assertEqual(
        returned_value,
        software_string
    )



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import fnmatch
import glob
import os
import re
import warnings

import pkg_resources
import requests
from six.moves.configparser import ConfigParser
import six

try:
  import subprocess32 as subprocess
except ImportError:
  import subprocess  # type: ignore
  subprocess  # pyflakes

try:
  from typing import  Dict, Iterable, List
except ImportError:
  pass

from ..slap.standalone import StandaloneSlapOS
from ..grid.utils import md5digest


def checkSoftware(slap, software_url):
  # type: (StandaloneSlapOS, str) -> None
  """Check software installation.

  This perform a few basic static checks for common problems
  with software installations.
  """
  # Check that all components set rpath correctly and we don't have miss linking any libraries.
  # Also check that they are not linked against system libraries, except a list of core
  # system libraries.
  system_lib_allowed_list = set((
      'libanl',
      'libc',
      'libcrypt',
      'libdl',
      'libgcc_s',
      'libgomp',
      'libm',
      'libmvec',
      'libnsl',
      'libpthread',
      'libthread_db',
      'libresolv',
      'librt',
      'libstdc++',
      'libutil',
  ))

  # Some libraries might also not be found statically but provided at run time.
  missing_lib_allowed_list = set((
      # references to liblibgolang.so from projects outside of pygolang are resolved at runtime for now.
      # https://github.com/mdavidsaver/setuptools_dso/issues/11#issuecomment-808258994
      'liblibgolang',
  ))

  # we also ignore a few patterns for part that are known to be binary distributions,
  # for which we generate LD_LIBRARY_PATH wrappers or we don't use directly.
  ignored_file_patterns = set((
      '*/parts/java-re*/*',
      '*/parts/firefox*/*',
      '*/parts/chromium-*/*',
      '*/parts/chromedriver*/*',
      '*/parts/libreoffice-bin/*',
      '*/parts/wkhtmltopdf/*',
      # R uses wrappers to relocate symbols
      '*/r-language/lib*/R/*',
      # pulp is a git checkout with some executables
      '*/pulp-repository.git/src/pulp/solverdir/cbc*',
      # nss is not a binary distribution, but for some reason it has invalid rpath, but it does
      # not seem to be a problem in our use cases.
      '*/parts/nss/*',
      # npm packages containing binaries
      '*/node_modules/phantomjs*/*',
      '*/grafana/tools/phantomjs/*',
      '*/node_modules/puppeteer/*',
      # left over of compilation failures
      '*/*__compile__/*',
      # build dir for packages built in-place
      '*/parts/wendelin.core/build/*',
      # the depot_tools package used to build Chromium installs some
      # Python libraries lacking an rpath; these are not actually used
      # by Chromium itself
      '*/.vpython-root/*',
  ))

  software_hash = md5digest(software_url)
  error_list = []
  warning_list = []

  ldd_so_resolved_re = re.compile(
      r'\t(?P<library_name>.*) => (?P<library_path>.*) \(0x')
  ldd_already_loaded_re = re.compile(r'\t(?P<library_name>.*) \(0x')
  ldd_not_found_re = re.compile(r'\t(?P<library_name>.*) => not found.*')

  class DynamicLibraryNotFound(Exception):
    """Exception raised when ldd cannot resolve a library.
    """
  def getLddOutput(path):
    # type: (str) -> Dict[str, str]
    """Parse ldd output on shared object/executable as `path` and returns a mapping
    of library paths or None when library is not found, keyed by library so name.

    Raises a `DynamicLibraryNotFound` if any dynamic library is not found.

    Special entries, like VDSO ( linux-vdso.so.1 ) or ELF interpreter
    ( /lib64/ld-linux-x86-64.so.2 ) are ignored.
    """
    libraries = {}  # type: Dict[str, str]
    try:
      ldd_output = subprocess.check_output(
          ('ldd', path),
          stderr=subprocess.STDOUT,
          universal_newlines=True,
      )
    except subprocess.CalledProcessError as e:
      if e.output not in ('\tnot a dynamic executable\n',):
        raise
      return libraries
    if ldd_output == '\tstatically linked\n':
      return libraries

    not_found = []
    for line in ldd_output.splitlines():
      resolved_so_match = ldd_so_resolved_re.match(line)
      ldd_already_loaded_match = ldd_already_loaded_re.match(line)
      not_found_match = ldd_not_found_re.match(line)
      if resolved_so_match:
        libraries[resolved_so_match.group(
            'library_name')] = resolved_so_match.group('library_path')
      elif ldd_already_loaded_match:
        # VDSO or ELF, ignore . See https://stackoverflow.com/a/35805410/7294664 for more about this
        pass
      elif not_found_match:
        library_name = not_found_match.group('library_name')
        if library_name.split('.')[0] not in missing_lib_allowed_list:
          not_found.append(line)
      else:
        raise RuntimeError('Unknown ldd line %s for %s.' % (line, path))
    if not_found:
      not_found_text = '\n'.join(not_found)
      raise DynamicLibraryNotFound(
          '{path} has some not found libraries:\n{not_found_text}'.format(
              **locals()))
    return libraries

  def checkExecutableLink(paths_to_check, valid_paths_for_libs):
    # type: (Iterable[str], Iterable[str]) -> List[str]
    """Check shared libraries linked with executables in `paths_to_check`.
    Only libraries from `valid_paths_for_libs` are accepted.
    Returns a list of error messages.
    """
    valid_paths_for_libs = [os.path.realpath(x) for x in valid_paths_for_libs]
    executable_link_error_list = []
    for path in paths_to_check:
      for root, dirs, files in os.walk(path):
        for f in files:
          f = os.path.join(root, f)
          if any(fnmatch.fnmatch(f, ignored_pattern)
                 for ignored_pattern in ignored_file_patterns):
            continue
          if os.access(f, os.X_OK) or fnmatch.fnmatch(f, '*.so'):
            try:
              libs = getLddOutput(f)
            except DynamicLibraryNotFound as e:
              executable_link_error_list.append(str(e))
            else:
              for lib, lib_path in libs.items():
                if lib.split('.')[0] in system_lib_allowed_list:
                  continue
                lib_path = os.path.realpath(lib_path)
                # dynamically linked programs can only be linked with libraries
                # present in software or in shared parts repository.
                if any(lib_path.startswith(valid_path)
                       for valid_path in valid_paths_for_libs):
                  continue
                executable_link_error_list.append(
                    '{f} uses system library {lib_path} for {lib}'.format(
                        **locals()))
    return executable_link_error_list

  software_directory = os.path.join(slap.software_directory, software_hash)
  paths_to_check = set((software_directory, ))

  # Compute the paths to check by inspecting buildout installed database
  # for this software. We are looking for shared parts installed by recipes.
  config_parser = ConfigParser()
  config_parser.read(os.path.join(software_directory, '.installed.cfg'))
  for section_name in config_parser.sections():
    for option_name in 'location', '__buildout_installed__':
      if config_parser.has_option(section_name, option_name):
        for section_path in config_parser.get(section_name, option_name).splitlines():
          if section_path and not section_path.startswith(software_directory):
            paths_to_check.add(section_path)

  error_list.extend(
      checkExecutableLink(
          paths_to_check,
          tuple(paths_to_check) + tuple(slap._shared_part_list),
      ))

  # check this software is not referenced in any shared parts.
  for signature_file in glob.glob(
      os.path.join(
          slap.shared_directory,
          '*',
          '*',
          '.buildout-shared.json',
      )):
    with open(signature_file) as f:
      signature_content = f.read()
    if software_hash in signature_content:
      error_list.append(
          "Shared part is referencing non shared part or software {}\n{}\n".format(
              signature_file, signature_content))

  def checkEggsVersionsKnownVulnerabilities(
      egg_directories,
      safety_db=requests.get(
          'https://raw.githubusercontent.com/pyupio/safety-db/master/data/insecure_full.json'
      ).json()):
    # type: (List[str], Dict) -> Iterable[str]
    """Check eggs against known vulnerabilities database from https://github.com/pyupio/safety-db
    """
    def get_python_versions():
      # () -> Iterable[str]
      """Returns all python versions used in egg_directories 
      """
      python_versions = set()
      for egg_dir in egg_directories:
        basename, _ = os.path.splitext(os.path.basename(egg_dir))
        match = pkg_resources.EGG_NAME(basename)
        if match:
          pyver = match.group('pyver')
          if pyver:
            python_versions.add(pyver)
      return python_versions

    for python in get_python_versions():
      env = pkg_resources.Environment(egg_directories, python=python)
      for egg in env:
        known_vulnerabilities = safety_db.get(egg)
        if known_vulnerabilities:
          for distribution in env[egg]:
            for known_vulnerability in known_vulnerabilities:
              for vulnerable_spec in known_vulnerability['specs']:
                for req in pkg_resources.parse_requirements(egg +
                                                            vulnerable_spec):
                  vulnerability_description = "\n".join(
                      u"{}: {}".format(*item)
                      for item in known_vulnerability.items())
                  if distribution in req:
                    yield (
                        u"{egg} use vulnerable version {distribution.version} because {vulnerable_spec}.\n"
                        "{vulnerability_description}\n".format(**locals()))

  warning_list.extend(
      checkEggsVersionsKnownVulnerabilities(
          glob.glob(
              os.path.join(
                  slap.software_directory,
                  software_hash,
                  'eggs',
                  '*',
              )) + glob.glob(
                  os.path.join(
                      slap.software_directory,
                      software_hash,
                      'develop-eggs',
                      '*',
                  ))))

  if warning_list:
    if six.PY2:
      # https://bugs.python.org/issue34752
      warnings.warn('\n'.join(warning_list).encode('utf-8'))
    else:
      warnings.warn('\n'.join(warning_list))
  if error_list:
    raise RuntimeError('\n'.join(error_list))
